# -*- coding: utf-8 -*-
"""
ReSTructured additional roles
"""
import os, re

from docutils import nodes, utils
from docutils.parsers.rst import roles

from django.conf import settings
from django.core.urlresolvers import reverse
from django.core.cache import cache
from django.contrib.sites.models import Site

from sveedocuments.models import Page, Attachment

_ATTACHMENT_ROLE_REGEX = re.compile(r"^(?:id)(?P<id>[0-9]+)(?:\-)(?P<slug>.*?)$")

def rst_parser_error(msg, rawtext, text, lineno, inliner):
        msg = inliner.reporter.error(msg, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

def get_page_slugs(force_update_cache=False):
    """
    Get a dict of all visible *Pages* as a tuple ``(slug, title)``
    
    Try to get it from the cache if it exist, else build it
    """
    if force_update_cache or not cache.get(settings.PAGE_SLUGS_CACHE_KEY_NAME):
        slugs_map = dict(Page.objects.filter(visible=True).values_list('slug', 'title'))
        cache.set(settings.PAGE_SLUGS_CACHE_KEY_NAME, slugs_map)
        return slugs_map
    return cache.get(settings.PAGE_SLUGS_CACHE_KEY_NAME)

def page_link(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Role to make a reference link to other *Pages* by using their ``slug``
    
    Usage in document :
    
        Blah blah :page:`my-page-slug`
    """
    # Get the page slugs map
    slugs = get_page_slugs()
    # Throw error if the given slug does not exist
    if text not in slugs and not settings.DOCUMENTS_PARSER_WIKIROLE_SILENT_WARNING:
        msg = inliner.reporter.error('Page with slug "%s" does not exist.' % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    # Add a class to the item
    options.update({'classes': ['documents_page_link']})
    roles.set_classes(options)
    # Return the node as reference to display the link for the given page's slug
    site_current = Site.objects.get_current()
    url = "http://{0}{1}".format(site_current.domain, reverse('documents-page-details', args=[text]))
    node = nodes.reference(rawtext, utils.unescape(slugs[text]), refuri=url, **options)
    return [node], []

roles.register_local_role('page', page_link)


def get_page_attachment_slugs(page_id, force_update_cache=False):
    """
    Get a dict of all Attachments linked to a Page
    
    Try to get it from the cache if it exist, else build it
    """
    cache_key = settings.PAGE_ATTACHMENTS_SLUGS_CACHE_KEY_NAME.format(page_id)
    if force_update_cache or not cache.get(cache_key):
        page = Page.objects.get(pk=page_id)
        slugs_map = dict(page.attachment.all().values_list('slug', 'file'))
        cache.set(cache_key, slugs_map)
        return slugs_map
    return cache.get(cache_key)

def page_attachment(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """
    Role to make a reference link to a Page's attachment
    
    Usage in document :
    
        Blah blah :attachment:`idX-slug`
        
    Where X is the page id and slug his slugname
    
    The page id is needed because i can't find a clean way to give some page context to 
    the docutils parser.
    """
    matched = _ATTACHMENT_ROLE_REGEX.match(text)
    if not matched or len(matched.groups())<2:
        return rst_parser_error('Attachment role syntax is not respected with "{0}", you should write something like "idXX-ATTACHMENT_SLUG".'.format(text), rawtext, text, lineno, inliner)
    
    # Get the page slugs map
    pk, attachment_slug = matched.groups()
    try:
        slugs_map = get_page_attachment_slugs(pk)
    except Page.DoesNotExist:
        return rst_parser_error('Page with id "{pk}" does not exist in pattern "{pattern}"'.format(pk=pk, pattern=text), rawtext, text, lineno, inliner)
    else:
        if attachment_slug not in slugs_map and not settings.DOCUMENTS_PARSER_WIKIROLE_SILENT_WARNING:
            return rst_parser_error('Attachment with slug "{slug}" does not exist for page id "{pk}" in pattern "{pattern}".'.format(pk=pk, slug=attachment_slug, pattern=text), rawtext, text, lineno, inliner)
        link = slugs_map[attachment_slug]
        # Add a class to the item
        options.update({'classes': ['documents_page_attachment']})
        roles.set_classes(options)
        # Return the node as reference to display the link for the given page's slug
        node = nodes.reference(rawtext, utils.unescape(attachment_slug), refuri=os.path.join(settings.MEDIA_URL, link), **options)
        return [node], []

roles.register_local_role('attachment', page_attachment)



# -*- coding: utf-8 -*-
"""
ReStructuredText parser facilities
"""
import copy

try:
    from xml.etree.cElementTree import fromstring, tostring, ElementTree
except ImportError:
    from xml.etree.ElementTree import fromstring, tostring, ElementTree

from django.utils.encoding import smart_str

import docutils
import docutils.core

from rstview.parser import get_functional_settings

import sveedocuments.utils.rest_roles

def extract_toc(source, setting_key="default", body_only=True, initial_header_level=None, silent=True):
    """
    Extract the TOC from the rendered content by the parser
    
    This is very tricky, we add the ``contents`` directive to the document, 
    parse it again (sic) with docutils, then parse it with ElementTree to find 
    the TOC element to extract it. We assume the first element is allways the TOC.
    
    DEPRECATED: for some documents this cause Etree parsing errors
    """
    parser_settings = get_functional_settings(setting_key, body_only, initial_header_level, silent)
    
    toc_internal_id = "private-page-toc-menu"
    source = u".. contents:: {tocid}\n\n{source}".format(tocid=toc_internal_id, source=source)
    parts = docutils.core.publish_parts(source=smart_str(source), writer_name="html4css1", settings_overrides=parser_settings)

    extracted = fromstring((u"<div id=\"document_root_body\">"+parts['fragment']+u"</div>").encode('utf-8')).find("div")
    if extracted and 'id' in extracted.keys() and extracted.get("id") == toc_internal_id:
        return tostring(extracted.find("ul"), encoding="UTF-8").replace("<?xml version='1.0' encoding='UTF-8'?>", '').strip()
    
    return ''



# -*- coding: utf-8 -*-
"""
Page's attachments management views
"""
from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404
from django.views import generic

from braces.views import PermissionRequiredMixin

from sveedocuments.models import ATTACHMENTS_WITH_SENDFILE, Page, Attachment
from sveedocuments.forms.attachment import AttachmentForm
from sveedocuments.views.board.page import PageTabsContentMixin
from sveedocuments.utils.braces_addons import DetailListAppendView, DirectDeleteView

class PageAttachmentsView(PermissionRequiredMixin, PageTabsContentMixin, DetailListAppendView):
    """
    Form view to add file attachments to a Page
    """
    model = Attachment
    form_class = AttachmentForm
    template_name = "sveedocuments/board/page_attachments.html"
    permission_required = 'sveedocuments.change_page'
    raise_exception = True
    context_parent_object_name = 'page_instance'
    
    def get_page_object(self):
        return self.get_parent_object()
    
    def get_parent_object(self):
        return get_object_or_404(Page, slug=self.kwargs['slug'])
    
    def get_queryset(self):
        return self.parent_object.attachment.all()

    def get_success_url(self):
        return reverse('sveedocuments:page-attachments', args=[self.parent_object.slug])
        
    def get_context_data(self, **kwargs):
        context = super(PageAttachmentsView, self).get_context_data(**kwargs)
        context.update({
            'ATTACHMENTS_WITH_SENDFILE': ATTACHMENTS_WITH_SENDFILE,
        })
        return context

    def get_form_kwargs(self):
        kwargs = super(PageAttachmentsView, self).get_form_kwargs()
        kwargs.update({'author': self.request.user})
        return kwargs


class PageAttachmentDeleteView(PermissionRequiredMixin, DirectDeleteView):
    """
    View to delete a *Page* document
    """
    model = Attachment
    permission_required = 'sveedocuments.change_page'
    raise_exception = True
    memoize_old_object = True
    _memoized_attr = ['id', 'slug', 'title', 'page']
    
    def get_page_object(self):
        return get_object_or_404(Page, slug=self.kwargs['slug'])

    def get_success_url(self):
        return reverse('sveedocuments:page-attachments', args=[self.old_object['page'].slug])



# -*- coding: utf-8 -*-
"""
Root url's map for application
"""
from django.conf.urls import *

from sveedocuments.models import ATTACHMENTS_WITH_SENDFILE
from sveedocuments.views.page import (
    HelpPageView, PageIndexView, PageDetailsView, 
    PageSourceView
)

urlpatterns = patterns('',
    url(r'^$', PageIndexView.as_view(), name='index'),
    
    (r'^board/', include('sveedocuments.urls_board')),
    
    url(r'^sitemap/$', PageIndexView.as_view(), name='index'),
    url(r'^help/$', HelpPageView.as_view(), name='help'),
    
    url(r'^(?P<slug>[-\w]+)/$', PageDetailsView.as_view(), name='page-details'),
    url(r'^(?P<slug>[-\w]+)/source/$', PageSourceView.as_view(), name='page-source'),
)

if ATTACHMENTS_WITH_SENDFILE:
    from sveedocuments.views.attachment import AttachmentProtectedDownloadView
    urlpatterns += patterns('',
        url(r'^(?P<slug>[-\w]+)/attachment/(?P<attachment_id>\d+)/$', AttachmentProtectedDownloadView.as_view(), name='page-attachment-download'),
    )



# -*- coding: utf-8 -*-
"""
sveedocuments default settings
"""
gettext = lambda s: s

# Forbidden words for slug values in documents
DOCUMENTS_PAGE_RESERVED_SLUGS = ('add','board','help','inserts','preview','sitemap')

# Default template to generate tree menu
DOCUMENTS_PAGE_TREEMENU = "sveedocuments/menu_tree.html"
# Default template to generate flat menu
DOCUMENTS_PAGE_FLATMENU = "sveedocuments/menu_flat.html"

# Enable or disable Pages archiving
DOCUMENTS_PAGE_ARCHIVED = True

# Page view restriction, if True page viewing will require to be authenticated, 
# if False page viewing is free for all
DOCUMENTS_PAGE_RESTRICTED = False

# To enable django-sendfile usage when DOCUMENTS_PAGE_RESTRICTED is True
DOCUMENTS_ATTACHMENT_USE_SENDFILE = True

# Available templates for Pages
DOCUMENTS_PAGE_TEMPLATES = {
    'default': ('sveedocuments/page_details/page_default.html', gettext('Default template with document content only')),
    # Disabled, because TOC extract is buggy
    #'with_toc': ('sveedocuments/page_details/page_with-toc.html', gettext('Template with a Table Of Content and the document content')),
}

DOCUMENTS_PAGE_TEMPLATE_DEFAULT = 'elastic'

# Active a silent reporter to avoid warnings about missing page on role ``:page:xxx``
# If ``True`` the role will be transformed to link despite the page does not exist, 
# if ``False`` the warning will be inserted in the render
DOCUMENTS_PARSER_WIKIROLE_SILENT_WARNING = False

"""
WARNING: Sample additional Django-CodeMirror settings, you have to put them yourself in your project settings
"""
#DJANGOCODEMIRROR_USER_SETTINGS_COOKIE_NAME = "djangocodemirror_user_settings"
#CODEMIRROR_SETTINGS = {
    #'sveetchies-documents-edit-page': {
        #'mode': 'rst',
        #'csrf': 'CSRFpass',
        #'preview_url': ('sveedocuments:preview',),
        #'quicksave_url': ('sveedocuments:page-quicksave',),
        #'quicksave_datas': 'DJANGOCODEMIRROR_OBJECT',
        #'lineWrapping': True,
        #'lineNumbers': True,
        #'search_enabled': True,
        #'settings_cookie': DJANGOCODEMIRROR_USER_SETTINGS_COOKIE_NAME,
        #'help_link': ('sveedocuments:help',),
        #'settings_url': ('sveedocuments:editor-settings', [], {}),
    #},
    #'sveetchies-documents-edit-insert': {
        #'mode': 'rst',
        #'csrf': 'CSRFpass',
        #'preview_url': ('sveedocuments:preview',),
        #'quicksave_url': ('sveedocuments:insert-quicksave',),
        #'quicksave_datas': 'DJANGOCODEMIRROR_OBJECT',
        #'lineWrapping': True,
        #'lineNumbers': True,
        #'search_enabled': True,
        #'settings_cookie': DJANGOCODEMIRROR_USER_SETTINGS_COOKIE_NAME,
        #'help_link': ('sveedocuments:help',),
        #'settings_url': ('sveedocuments:editor-settings', [], {}),
    #},
#}
#CODEMIRROR_SETTINGS['sveetchies-documents-add-page'] = CODEMIRROR_SETTINGS['sveetchies-documents-edit-page'].copy()
#CODEMIRROR_SETTINGS['sveetchies-documents-add-page']['quicksave_url'] = None
#CODEMIRROR_SETTINGS['sveetchies-documents-add-insert'] = CODEMIRROR_SETTINGS['sveetchies-documents-edit-insert'].copy()
#CODEMIRROR_SETTINGS['sveetchies-documents-add-insert']['quicksave_url'] = None

# Add custom cache keys to delete with the command option ``django-admin documents --clearcache``
DOCUMENTS_CACHE_KEYS_TO_CLEAN = []

"""
Internal settings only, can't be overriden from your webapp settings
"""

# Cache keys for document elements
PAGE_RENDER_CACHE_KEY_NAME = 'documents-render-page_{id}-setting_{setting}'
PAGEREV_RENDER_CACHE_KEY_NAME = 'documents-render-page-revision_{id}-setting_{setting}'
INSERT_RENDER_CACHE_KEY_NAME = 'documents-render-insert_{id}-setting_{setting}-hlv_{header_level}'
PAGE_TOC_CACHE_KEY_NAME = 'documents-toc-page_{id}-setting_{setting}'
PAGEREV_TOC_CACHE_KEY_NAME = 'documents-toc-page-revision_{id}-setting_{setting}'
INSERT_TOC_CACHE_KEY_NAME = 'documents-toc-insert_{id}-setting_{setting}-hlv_{header_level}'
PAGE_SLUGS_CACHE_KEY_NAME = 'documents-page_slugs'
PAGE_ATTACHMENTS_SLUGS_CACHE_KEY_NAME = 'documents-page-attachments-slugs'



# -*- coding: utf-8 -*-
"""
Page document views
"""
import json, os
from datetime import datetime
from cStringIO import StringIO

from django.conf import settings
from django.core.urlresolvers import reverse
from django.http import HttpResponse, Http404
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.views import generic
from django.views.generic.detail import SingleObjectMixin

from mptt.templatetags.mptt_tags import cache_tree_children

from braces.views import LoginRequiredMixin, PermissionRequiredMixin

from djangocodemirror.views import SampleQuicksaveMixin

from rstview.parser import SourceParser

from sveedocuments import __init__ as sveedocuments_basemod
from sveedocuments.models import ATTACHMENTS_WITH_SENDFILE, Page, Attachment
from sveedocuments.forms.page import PageForm, PageEditForm, PageQuickForm
from sveedocuments.forms.attachment import AttachmentForm
from sveedocuments.utils.objects import get_instance_children
from sveedocuments.utils.braces_addons import DetailListAppendView, DirectDeleteView, DownloadMixin

class PageIndexView(generic.TemplateView):
    """
    Pages index
    
    TODO: Sitemap view should be conditionnated on DOCUMENTS_PAGE_RESTRICTED like PageDetailsView
    """
    template_name = "sveedocuments/index.html"
    
    def get(self, request, *args, **kwargs):
        # We can't force queryset evaluation here else it will mess the tree 
        # resolution from mptt
        page_list = Page.objects.filter(visible=True)
        
        context = {
            'page_list': page_list,
            'json_tree': [],
        }
        
        # Fill the JSON tree if there is at least one entry
        if page_list:
            context['json_tree'] = json.dumps(self.get_recursed_tree( cache_tree_children(page_list) ))
            
        return self.render_to_response(context)
    
    def get_recursed_tree(self, root_nodes):
        """
        Get the recursed tree from the queryset
        """
        nodes = []
        for n in root_nodes:
            nodes.append({
                "id": n.pk,
                "label": n.title,
                "slug": n.slug,
                "view_url": reverse('sveedocuments:page-details', args=[n.slug]),
                "children": self.get_recursed_tree(n.get_children())
            })
        return nodes

class HelpPageView(generic.TemplateView):
    """
    Help document
    
    TODO: Should be conditionnated on DOCUMENTS_PAGE_RESTRICTED like PageDetailsView
    """
    template_name = "sveedocuments/help.html"
    
    def get(self, request, *args, **kwargs):
        path_root = os.path.abspath(os.path.dirname(sveedocuments_basemod.__file__))
        f = open(os.path.join(path_root, "HELP.rst"))
        content = f.read()
        f.close()
        
        context = {'content' : SourceParser(content, silent=False)}
        return self.render_to_response(context)


class PageDetailsMixin(object):
    """
    *Page* view
    """
    model = Page
    context_object_name = "page_instance"
    template_name = "sveedocuments/page_details/page_default.html"
    
    def get_object(self, *args, **kwargs):
        """
        Memorize object to avoid multiple database access when using ``get_object()`` 
        method
        """
        cache_key = "_cache_get_object"
        if not hasattr(self, cache_key):
            setattr(self, cache_key, super(PageDetailsMixin, self).get_object(*args, **kwargs))
        return getattr(self, cache_key)
    
    def get_attachments(self):
        return self.object.attachment.all()
        
    def get_context_data(self, **kwargs):
        context = super(PageDetailsMixin, self).get_context_data(**kwargs)
        context.update({
            'attachments': self.get_attachments(),
            'ATTACHMENTS_WITH_SENDFILE': ATTACHMENTS_WITH_SENDFILE,
        })
        return context
    
    def get(self, request, **kwargs):
        # Check if the object is ``visible``
        if not self.get_object().visible:
            raise Http404
        return super(PageDetailsMixin, self).get(request, **kwargs)
    
    def get_template_names(self):
        return [self.object.get_template()]

if settings.DOCUMENTS_PAGE_RESTRICTED:
    class PageDetailsView(PageDetailsMixin, LoginRequiredMixin, generic.DetailView):
        pass
else:
    class PageDetailsView(PageDetailsMixin, generic.DetailView):
        pass


class PageSourceView(PageDetailsView):
    """
    Raw content *Page* view
    """
    def get(self, request, *args, **kwargs):
        # Check if the object is ``visible``
        if not self.get_object().visible:
            raise Http404
        return HttpResponse(self.get_object().content, content_type="text/plain; charset=utf-8")



# -*- coding: utf-8 -*-
"""
Page's attachment views
"""
import os

from django.conf import settings
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.views import generic

from sendfile import sendfile

from braces.views import LoginRequiredMixin

from sveedocuments.models import Attachment
from sveedocuments.views.page import PageDetailsMixin

class AttachmentProtectedDownloadView(PageDetailsMixin, LoginRequiredMixin, generic.DetailView):
    """
    View to download protected attachment
    """
    def get_attachment(self):
        return get_object_or_404(Attachment, page=self.object, pk=self.kwargs['attachment_id'])
    
    def get(self, request, **kwargs):
        self.object = self.get_object()
        self.attachment = self.get_attachment()
        
        # If page is not visible, their attachments are not available to download
        if not self.object.visible:
            raise Http404
        
        file_path = os.path.join(settings.PROJECT_PATH, self.attachment.file.path)
        return sendfile(request, file_path, attachment=True, attachment_filename=os.path.basename(file_path))



# -*- coding: utf-8 -*-
"""
Templates tags divers pour les documents
"""
from django.conf import settings
from django import template
from django.utils.safestring import mark_safe

from sveedocuments.models import Insert, Page
from sveedocuments.utils.templatetags import resolve_string_or_variable

register = template.Library()

class PageMenuTagRender(template.Node):
    """
    Generate HTML of the node *PageMenuTagRender*
    
    TODO: @flat_mode usage is a failure because it works like the tree, flat menu should 
          never include the Page from where to start and display only his direct 
          children.
    """
    def __init__(self, page_var_name, template_path_varname=None, flat_mode=0):
        """
        :type page_var_name: string or object ``django.db.models.Model``
        :param page_var_name: Nom de variable de l'instance ou un string 
                                        pour le slug.
                                        
        :type template_path_varname: string
        :param template_path_varname: (optionnel) Chemin d'un template à utiliser pour 
                                      le rendu du menu à générer
                                        
        :type flat_mode: int
        :param flat_mode: (optionnel) Indique si on doit gérer un menu avec seulement les 
                          pages adjacentes (flat) à la page ciblée ou bien l'arborescence 
                          récursive de ses enfants (tree).
        """
        self.page_var_name = page_var_name
        self.template_path_varname = template_path_varname
        self.flat_mode = flat_mode
    
    def render(self, context):
        """
        Rendu de tout les éléments du formulaire indiqué
        
        :type context: object ``django.template.Context``
        :param context: Objet du contexte du tag.
        
        :rtype: string
        :return: Le rendu généré pour le tag capturé.
        """
        html = ''
        parser_kwargs = {}
        active_page_instance = None
        
        # Résolution des arguments
        page_var = resolve_string_or_variable(self.page_var_name, context)
        # Template par défaut selon le mode (arborescence/plat)
        if not self.flat_mode:
            self.template_path = settings.DOCUMENTS_PAGE_TREEMENU
        else:
            self.template_path = settings.DOCUMENTS_PAGE_FLATMENU
        # Custom template if any
        if self.template_path_varname:
            self.template_path = resolve_string_or_variable(self.template_path_varname, context)
        
        # Transmet au contexte du tag l'instance de la page courante si elle est présente 
        # dans le contexte de la page
        if 'page_instance' in context:
            active_page_instance = context['page_instance']
        
        return self.build(context, page_var, active_page_instance)
    
    def build(self, context, page_var, active_page_instance=None):
        """
        Calcul du rendu du menu, renvoi le menu html ou bien une chaine vide dans 
        certains cas ou la cible n'existe pas
        """
        # Recherche d'un zéro pour signifier qu'on doit commencer non pas depuis une 
        # instance mais depuis la racine de l'arborescence
        if isinstance(page_var, int):
            if page_var == 0:
                page_instance = None
                page_list = Page.objects
                if self.flat_mode:
                    page_list = page_list.root_nodes()
                page_list = page_list.filter(visible=True)
            # Page ids are not supported
            else:
                return ''
        # Instance directement transmise
        elif isinstance(page_var, Page):
            page_instance = page_var
            if self.flat_mode == 1:
                page_list = page_var.get_siblings(include_self=True).filter(visible=True)
            elif self.flat_mode == 2:
                page_list = page_var.get_children().filter(visible=True)
            else:
                page_list = page_var.get_descendants(include_self=False).filter(visible=True)
        # Réception d'un slug pour récupérer l'instance
        else:
            try:
                instance = Page.objects.get(slug=page_var, visible=True)
            except Page.DoesNotExist:
                return ''
            else:
                page_instance = instance
                if self.flat_mode == 1:
                    page_list = instance.get_siblings(include_self=True).filter(visible=True)
                elif self.flat_mode == 2:
                    page_list = instance.get_children().filter(visible=True)
                else:
                    page_list = instance.get_descendants(include_self=False).filter(visible=True)
        
        # Aucun élément, on ne poursuit pas
        if not page_list:
            return ''
        
        # Permet à un menu "plat" (flat) de retrouver le parent de la page en cours
        # NOTE: ne marchera qu'avec un menu à la racine, il faut utiliser get_ancestors 
        # et test "in"
        if active_page_instance:
            active_page_instance.root_slug = None
            if self.flat_mode:
                active_page_instance.root_slug = active_page_instance.get_root().slug
        
        subcontext = {
            'active_page_instance': active_page_instance,
            'page_instance': page_instance,
            'page_list': page_list,
        }
        
        html = template.loader.get_template(self.template_path).render(template.Context(subcontext))
        
        return mark_safe(html)

@register.tag(name="document_page_treemenu")
def do_document_page_treemenu(parser, token):
    """
    Display a tree menu of **Pages**
    
    TODO: Pouvoir limiter le niveau de récursivité (0->n)
    
    Arguments :
    
    document_object
        Instance "Page" ou un string contenant un slug de l'instance à récupérer
        
    Exemple d'utilisation par une instance Page : ::
    
        {% document_page_treemenu instance %}
        
    Ou par le slug de l'Insert à utiliser : ::
    
        {% document_page_treemenu "my-page-slug" %}
    
    Ou la totalité de l'arborescence des pages avec un zéro pour signifier qu'on 
    commence depuis la racine : ::
    
        {% document_page_treemenu 0 %}
    
    On peut aussi spécifier un template autre que celui par défaut : ::
    
        {% document_page_treemenu instance "mymenu.html" %}
        
    :type parser: object ``django.template.Parser``
    :param parser: Objet du parser de template.
    
    :type token: object ``django.template.Token``
    :param token: Objet de la chaîne découpée du tag capturé dans le template.
    
    :rtype: object ``PageMenuTagRender``
    :return: L'objet du générateur de rendu du tag.
    """
    args = token.split_contents()
    if len(args) < 2:
        raise template.TemplateSyntaxError, "You need to specify at less a \"Page\" instance, a slug or a 0 (for full sitemap)"
    else:
        return PageMenuTagRender(*args[1:])

do_document_page_treemenu.is_safe = True

@register.tag(name="document_page_flat_adjacents")
def do_document_page_flat_adjacents(parser, token):
    """
    Display a flat menu of adjacents **Pages**
    """
    args = token.split_contents()
    if len(args) < 2:
        raise template.TemplateSyntaxError, "You need to specify at less a \"Page\" instance or a slug"
    else:
        return PageMenuTagRender(*args[1:], flat_mode=1)

do_document_page_flat_adjacents.is_safe = True

@register.tag(name="document_page_flat_children")
def do_document_page_flat_children(parser, token):
    """
    Display a flat menu of children **Pages**
    """
    args = token.split_contents()
    if len(args) < 2:
        raise template.TemplateSyntaxError, "You need to specify at less a \"Page\" instance or a slug"
    else:
        return PageMenuTagRender(*args[1:], flat_mode=2)

do_document_page_flat_children.is_safe = True

@register.simple_tag
def pprint_recurse(elements):
    return recurse_list(elements)

def recurse_list(relations, sig=0):
    """
    Renvoi une liste à puce récursive de toute les relations
    """
    output = ""
    if not relations:
        return output
    for name, object_title, followed in relations:
        children = recurse_list(followed, sig=sig+1)
        output += "<li><p><strong>%s :</strong> %s</p>%s</li>" % (name.title(), object_title, children)
    return ("""<ul class="level_%s">"""%sig) + output + "</ul>"



# -*- coding: utf-8 -*-
"""
Utilitaires liés aux champs de fichiers dans les modèles de données
"""
from datetime import datetime as real_datetime
from os import makedirs
from os.path import basename, dirname, exists, isdir, join, splitext

from django.conf import settings
from django.utils.datetime_safe import strftime

def content_file_name(upload_to_path, instance, filename=None, new_extension=None):
    """
    Méthode à utiliser pour l'attribut "upload_to" d'un FileField/ImageField pour 
    renommer automatiquement le nom de fichier avec un nom unique et empêcher la 
    conservation de fichier avec des caractères spéciaux.
    
    À ne pas utiliser directement sur l'attribut "FileField.upload_to" 
    mais via un callable pour indiquer le chemin "upload_to" par exemple :
    
        ATTACH_FILE_UPLOADTO = lambda x,y: content_file_name('attach/%Y/%m/%d', x, y)
    
    puis
    
        file = models.FileField(u'fichier', upload_to=ATTACH_FILE_UPLOADTO)
    """
    upload_to = real_datetime.now().strftime(upload_to_path)
    if not filename:
        return upload_to
    filename = get_unique_filename(filename, new_extension=new_extension)
    return '/'.join([upload_to, filename])

def get_unique_filename(filename, new_filename=None, new_extension=None):
    """
    Génère un nouveau nom pour un fichier en gardant son extension
    
    Soit le nouveau nom est généré à partir de la date 
    (heures+minutes+secondes+microsecondes) soit un nouveau nom est spécifié et on 
    l'utilise tel quel.
    
    :type filename: string
    :param filename: Nom du fichier original
    
    :type new_filename: string
    :param new_filename: (optional) Nouveau nom de fichier personnalisé, ceci implique que 
                         la date ne sera pas insérée dans le nouveau nom
    
    :type new_extension: string
    :param new_extension: (optional) Force une nouvelle extension de fichier au lieu de 
                          celle de l'original. À spécifier sans le "point" de début d'une 
                          extension.
    
    :rtype: string
    :return: Nouveau nom de fichier
    """
    if new_extension:
        extension = new_extension
    else:
        extension = splitext(filename)[1][1:]
    if not new_filename:
        now = real_datetime.now()
        return '%s%s%s%s.%s' % (unicode(now.hour).zfill(2), unicode(now.minute).zfill(2), unicode(now.second).zfill(2), unicode(now.microsecond).zfill(6), extension)
    else:
        return '%s.%s' % (new_filename, extension)

def generate_valid_path(filename, path, create_it=True, root_base=settings.MEDIA_ROOT, permissions=settings.FILE_UPLOAD_PERMISSIONS):
    """
    Génére le chemin prévu pour un fichier et créé le chemin si besoin
    
    :type filename: string
    :param filename: Nom du fichier cible
    
    :type path: string
    :param path: Chemin relatif (au répertoire des médias, ``settings.MEDIA_ROOT``) du 
                 répertoire contenant le fichier, peut contenir les paramètres de 
                 substitution possibles avec la méthode 
                 ``django.utils.datetime_safe.strftime``.
    
    :type create_it: bool
    :param create_it: (optional) Déclenche la création du répertoire si il n'existe pas déja
    
    :type root_base: string
    :param root_base: (optional) Chemin de base ou créer le fichier, par défaut dans 
                      le ``MEDIA_ROOT`` du projet.
    
    :type permissions: int
    :param permissions: (optional) Permission chmod sur le chemin à créer, par 
                             défaut utilise le ``FILE_UPLOAD_PERMISSIONS`` des settings 
                             du projet. Et si cette variable n'est pas renseigné dans vos 
                             settings, alors un mask 0744 sera utilisé.
    
    :rtype: tuple
    :return: * Chemin relatif du fichier
             * Chemin absolu du fichier
    """
    if not permissions:
        permissions = 0744
    # Forme le chemin absolu vers le nouvo fichier créé
    relative_path = strftime(real_datetime.now(), path)
    absolute_path = join(root_base, relative_path)
    # Création du répertoire
    if create_it:
        if not exists( absolute_path ):
            makedirs(absolute_path, permissions)
        elif not isdir( absolute_path ):
            # Le chemin existe et n'est pas un répertoire
            raise IOError("'%s' exists and is not a directory." % absolute_path)
    
    return join(relative_path, filename), join(absolute_path, filename)



# -*- coding: utf-8 -*-
"""
Pages management views
"""
from datetime import datetime

from django.conf import settings
from django.core.urlresolvers import reverse
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.views import generic
from django.views.generic.detail import SingleObjectMixin

from mptt.templatetags.mptt_tags import cache_tree_children

from braces.views import LoginRequiredMixin, PermissionRequiredMixin

from djangocodemirror.views import SampleQuicksaveMixin

from sveedocuments.models import ATTACHMENTS_WITH_SENDFILE, Page, Attachment
from sveedocuments.forms.page import PageForm, PageEditForm, PageQuickForm
from sveedocuments.forms.attachment import AttachmentForm
from sveedocuments.utils.objects import get_instance_children
from sveedocuments.utils.braces_addons import DetailListAppendView, DirectDeleteView, DownloadMixin


class PageTabsContentMixin(object):
    def get_context_data(self, **kwargs):
        context = super(PageTabsContentMixin, self).get_context_data(**kwargs)
        # Little trick to work with SingleObjectMixin and DetailListAppendView
        if hasattr(self, 'parent_object'):
            obj = self.parent_object
        else:
            obj = self.object
        
        context.update({
            'revisions_count': obj.revision.all().count()+1,
            'files_count': obj.attachment.all().count(),
        })
        return context


class PageCreateView(PermissionRequiredMixin, generic.CreateView):
    """
    Form view to create a *Page* document
    """
    model = Page
    context_object_name = "page_instance"
    template_name = "sveedocuments/board/page_form.html"
    form_class = PageForm
    permission_required = "sveedocuments.add_page"
    raise_exception = True
    _redirect_to_self = False

    def get(self, request, *args, **kwargs):
        self.parent_page_instance = self._get_parent(**kwargs)
        return super(PageCreateView, self).get(request, *args, **kwargs)
        
    def post(self, request, *args, **kwargs):
        self.parent_page_instance = self._get_parent(**kwargs)
        # Mark to go back to the form page after save, triggered by special submit
        if request.POST:
            if request.POST.get('submit_and_continue', False):
                self._redirect_to_self = True
        return super(PageCreateView, self).post(request, *args, **kwargs)
        
    def get_success_url(self):
        if self._redirect_to_self:
            return reverse('sveedocuments:page-edit', args=[self.object.slug])
        return reverse('sveedocuments:page-index')

    def _get_parent(self, **kwargs):
        if 'slug' in kwargs:
            return get_object_or_404(Page, slug=kwargs['slug'])
        return None
        
    def get_context_data(self, **kwargs):
        context = super(PageCreateView, self).get_context_data(**kwargs)
        context.update({
            'parent_page_instance': self.parent_page_instance,
        })
        return context
    
    def get_form_kwargs(self):
        kwargs = super(PageCreateView, self).get_form_kwargs()
        kwargs.update({
            'author': self.request.user,
            'parent': self.parent_page_instance,
            'initial': {'published': datetime.today()},
        })
        return kwargs


class PageEditView(PermissionRequiredMixin, PageTabsContentMixin, generic.UpdateView):
    """
    Form view to edit a *Page* document
    """
    model = Page
    context_object_name = "page_instance"
    template_name = "sveedocuments/board/page_form.html"
    form_class = PageEditForm
    permission_required = "sveedocuments.change_page"
    raise_exception = True
    _redirect_to_self = False
        
    def post(self, request, *args, **kwargs):
        # Mark to go back to the form page after save, triggered by special submit
        if request.POST:
            if request.POST.get('submit_and_continue', False):
                self._redirect_to_self = True
        return super(PageEditView, self).post(request, *args, **kwargs)
    
    def get_object(self, *args, **kwargs):
        """
        Forcing a new empty comment field in edit mode
        """
        obj = super(PageEditView, self).get_object(*args, **kwargs)
        obj.comment = ''
        return obj

    def get_success_url(self):
        if self._redirect_to_self:
            return reverse('sveedocuments:page-edit', args=[self.object.slug])
        return reverse('sveedocuments:page-index')

    def get_form_kwargs(self):
        kwargs = super(PageEditView, self).get_form_kwargs()
        kwargs.update({'author': self.request.user})
        return kwargs


class PageDeleteView(PermissionRequiredMixin, generic.DeleteView):
    """
    Form view to delete a *Page* document
    
    This display a tree of relations to confirm the object and relations deletion
    """
    model = Page
    context_object_name = "page_instance"
    template_name = "sveedocuments/board/page_delete.html"
    permission_required = "sveedocuments.delete_page"
    raise_exception = True

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)
        context.update({'relations': get_instance_children(self.object)})
        return self.render_to_response(context)

    def get_success_url(self):
        return reverse('sveedocuments:page-index')


class PageHistoryView(PermissionRequiredMixin, PageTabsContentMixin, generic.DetailView):
    """
    *Page* history
    """
    model = Page
    context_object_name = "page_instance"
    template_name = "sveedocuments/board/page_history.html"
    permission_required = "sveedocuments.change_page"
    raise_exception = True
    
    def get_object(self, *args, **kwargs):
        """
        Memorize object to avoid multiple database access when using ``get_object()`` 
        method
        """
        cache_key = "_cache_get_object"
        if not hasattr(self, cache_key):
            setattr(self, cache_key, super(PageHistoryView, self).get_object(*args, **kwargs))
        return getattr(self, cache_key)
        
    def get_context_data(self, **kwargs):
        context = super(PageHistoryView, self).get_context_data(**kwargs)
        context.update({
            'last_revisions': self.object.revision.all().order_by('-created'),
        })
        return context


class PageQuicksaveView(SampleQuicksaveMixin, PageEditView):
    """
    Quicksave view for a *Page* content
    """
    form_class = PageQuickForm
    
    def get_object(self, queryset=None):
        if self.request.POST.get('slug', False):
            self.kwargs['slug'] = self.request.POST['slug']
        return super(PageQuicksaveView, self).get_object(queryset=queryset)



# -*- coding: utf-8 -*-
"""
Model admin
"""
from django.contrib import admin
from models import *

class InsertAdmin(admin.ModelAdmin):
    list_display = ('slug', 'title', 'modified', 'author', 'visible' )
    list_filter = ('created', 'modified', 'visible')
    ordering = ('slug',)
    search_fields = ('title','content',)
    readonly_fields = ('author',)
    list_display_links = ('title',)
    prepopulated_fields = {"slug": ("title",)}
    
    def save_model(self, request, obj, form, change):
        """
        Automatically add/update the author field with the ``request.user``
        """
        instance = form.save(commit=False)
        instance.author = request.user
        instance.save()
        form.save_m2m()

        return instance

class PageAdmin(admin.ModelAdmin):
    list_display = ('title', 'order', 'modified', 'published', 'author', 'visible' )
    list_filter = ('created', 'modified', 'published', 'visible')
    ordering = ('order',)
    search_fields = ('title','content',)
    readonly_fields = ('author',)
    list_display_links = ('title',)
    prepopulated_fields = {"slug": ("title",)}
    
    def save_model(self, request, obj, form, change):
        """
        Automatically add/update the author field with the ``request.user``
        """
        instance = form.save(commit=False)
        instance.author = request.user
        instance.save()
        form.save_m2m()

        return instance

class PageRevisionAdmin(admin.ModelAdmin):
    list_display = ('title', 'revision', 'author', 'visible' )
    list_filter = ('created', 'published', 'visible')
    ordering = ('order', 'revision')
    search_fields = ('title','content',)
    readonly_fields = ('author', 'revision')
    list_display_links = ('title',)
    
    def save_model(self, request, obj, form, change):
        """
        Automatically add/update the author field with the ``request.user``
        """
        instance = form.save(commit=False)
        instance.author = request.user
        instance.save()
        form.save_m2m()

        return instance

admin.site.register(Insert, InsertAdmin)
admin.site.register(Page, PageAdmin)
admin.site.register(PageRevision, PageRevisionAdmin)
admin.site.register(Attachment)



# -*- coding: utf-8 -*-
"""
General Command line tool
"""
import StringIO

from optparse import OptionValueError, make_option

from django.conf import settings
from django.core.cache import cache
from django.core.management.base import CommandError, BaseCommand

from sveedocuments.models import Page, Insert

class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option("--clearcache", dest="clearcache", action="store_true", default=False, help="Clear all documents (Page and Insert) cache."),
        make_option("--treefix", dest="treefix", action="store_true", default=False, help="Rebuild the Pages tree. This is currently used to fix a bug when deleting some item. This should be a temporary trick that will be deleted when a correct fix will be finded."),
    )
    help = "General command for Sveetchies-documents"

    def handle(self, *args, **options):
        if len(args) != 0:
            raise CommandError("Command doesn't accept any arguments")
        
        self.clearcache = options.get('clearcache')
        self.treefix = options.get('treefix')
        self.verbosity = int(options.get('verbosity'))
        
        if self.clearcache:
            self.do_clearcache()
        
        if self.treefix:
            self.do_treefix()

    def do_treefix(self):
        """
        Rebuild Pages tree info
        """
        Page.tree.rebuild()

    def do_clearcache(self):
        """
        Clear all possible caches from documents
        """
        inserts = Insert.objects.all().order_by('id')
        for instance_item in inserts:
            keys = instance_item.clear_cache()
            
        pages = Page.objects.all().order_by('id')
        for page_item in pages:
            keys = page_item.clear_cache()
        
        if settings.DOCUMENTS_CACHE_KEYS_TO_CLEAN:
            cache.delete_many(settings.DOCUMENTS_CACHE_KEYS_TO_CLEAN)
            
        if self.verbosity:
            print "* All documents cache cleared"



# -*- coding: utf-8 -*-
"""
Parser template tags 
"""
from django import template
from django.conf import settings
from django.utils.encoding import force_unicode
from django.utils.safestring import mark_safe

from rstview.parser import SourceParser

from sveedocuments.templatetags import get_render_with_cache, get_toc_with_cache

register = template.Library()

def source_render(source, setting_key="default"):
    """
    Return the parser result from the given string source and settings
    """
    return mark_safe( SourceParser(source, setting_key=setting_key) )
source_render.is_safe = True
register.filter(source_render)

def document_render(document_instance, setting_key="default"):
    """
    Return the parser result from the content of the given instance (*Pages* or 
    *Insert*) and settings
    """
    return mark_safe( get_render_with_cache(document_instance, setting_key=setting_key) )
document_render.is_safe = True
register.filter(document_render)

def document_toc(document_instance, setting_key="default"):
    """
    Get the TOC (Table Of Content) from the content of the given instance (*Pages* 
    or *Insert*) and settings
    """
    return mark_safe( get_toc_with_cache(document_instance, setting_key=setting_key) )
document_toc.is_safe = True
register.filter(document_toc)



# -*- coding: utf-8 -*-
"""
Export/Import Command line tool

TODO: * Better command name;
      * Option to select only some items from entries;
      * Item fields to edit some other object field;

Use this tool with a JSON config that contain mapping between Pages object and files.

Export
------

This will fetch items from the database and create it as files (or rewrite them if exists).

If an item filepath directory destination does not exists this will raise a CommandError. 

Use '--smart' option to automatically create missing directories. Take care before that your 
filepath destination is correct.

Import
------

This will read items filepath and import their content into their Pages.

If a slug does not exists this will raise a CommandError.

Use '--smart' option to automatically create a new Page if it does not exist yet. It will use the slug as the title, all other fields are let at their default value except the 'author'.

Pages are created or edited with the given 'author' if any, else with the default 
author (the first superuser). And the 'comment' is something like 
"Imported from 'FILE'".

Sample :

[
    {
        "kind": "export",
        "name": "Exporting project README's",
        "items": [
            {
                "slug": "djangosveetchies",
                "filepath": "./README.2.rst"
            },
            ...
        ]
    },
    {
        "kind": "import",
        "author": "sveetch",
        "name": "Importing project README's",
        "items": [
            {
                "slug": "djangosveetchies",
                "filepath": "./README.rst"
            },
            ...
        ]
    },
    ...
]

"""
import os, json, StringIO

from optparse import OptionValueError, make_option

from django.conf import settings
from django.core.management.base import CommandError, BaseCommand
from django.contrib.auth.models import User

from sveedocuments.models import Page

class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option("--config", dest="config_file", default=None, help="JSON config file", metavar="FILEPATH"),
        make_option("--smart", dest="smart", action="store_true", default=False, help="If used, path destination that does not exist will be created if needed in Export mode, in Import mode this will create entry that does not allready exist. Be carefull, if your filepath entry is incorrect, this can do some mess on your Filesystem or Database."),
        make_option("--dry-run", dest="dry_run", action="store_true", default=False, help="Does not create any file or edit database, just do all the process"),
    )
    help = "Import/export command line tool"

    def handle(self, *args, **options):
        if len(args) != 0:
            raise CommandError("Command doesn't accept any arguments")
        
        config_file = options.get('config_file')
        self.verbosity = int(options.get('verbosity'))
        self.dry_run = int(options.get('dry_run'))
        self.smart = int(options.get('smart'))
        
        if not config_file:
            raise CommandError("You need to specify a config file")
        
        self.configs = self.parse_config_file(config_file)
        self.proceed()

    def parse_config_file(self, config_file):
        """
        Open the config file, validate it, fill in some useful datas and return it
        """
        items = []
        default = {
            "name": None,
            "prepend_path": None,
            "backup": False,
            "author": None,
        }
        
        # Default author is the first superuser finded
        self.default_author = User.objects.filter(is_superuser=True)[0]
        
        _configs = json.load(open(config_file))
        
        for i, entry in enumerate(_configs, start=1):
            if "name" not in entry or not entry.get("name"):
                entry["name"] = "Entry {0}".format(i)
            if "kind" not in entry or entry.get("kind") not in ["export", "import"]:
                raise CommandError("{0} : Missing a correct value for the 'kind' key".format(entry["name"]))
            if "author" in entry and entry.get("kind"):
                try:
                    entry["author"] = User.objects.get(username=entry["author"])
                except User.DoesNotExist:
                    raise CommandError("{0} : 'author' key : username '{1}' does not exist".format(entry["name"], entry["author"]))
            
            for j, item in enumerate(entry["items"], start=1):
                # Required keys
                if "slug" not in item or "filepath" not in item:
                    raise CommandError("{0} : Item #{1} must have 'slug' and 'filepath' keys.".format(entry["name"], j))
                # Check the page object
                item["object"] = None
                try:
                    item["object"] = Page.objects.get(slug=item["slug"])
                except Page.DoesNotExist:
                    if entry["kind"]=='export' or (entry["kind"]=='import' and not self.smart):
                        raise CommandError("{0} : Item #{1} : slug '{2}' does not exist".format(entry["name"], j, item["slug"]))
                        
                # Check filepath
                item["filepath"] = os.path.abspath(item["filepath"])
                item["_filepath_dir"], item["_filepath_filename"] = os.path.split(item["filepath"])
                if not self.smart and not os.path.exists(item["_filepath_dir"]):
                    raise CommandError("{0} : Item #{1} : filepath directory '{2}' does not exist, use --smart if you want to automatically create it.".format(entry["name"], j, item["_filepath_dir"]))
                if entry["kind"]=='import' and not os.path.exists(item["filepath"]):
                    raise CommandError("{0} : Item #{1} : filepath '{2}' does not exist.".format(entry["name"], j, item["filepath"]))
            
            _cfg = default.copy()
            _cfg.update(entry)
            items.append(_cfg)
        
        return items

    def proceed(self):
        """
        Proceed to the entry action 
        """
        for i, entry in enumerate(self.configs, start=1):
            if hasattr(self, "do_{0}".format(entry["kind"])):
                getattr(self, "do_{0}".format(entry["kind"]))(entry)
            else:
                raise CommandError("{0} : Command method for kind '{1}' does not exist.".format(entry["name"], entry["kind"]))

    def do_import(self, entry):
        """
        Import documents from their file destination to their database object
        """
        if self.verbosity > 0:
            print "=== Import '{0}' ===".format(entry["name"])
        
        for i, item in enumerate(entry['items'], start=1):
            # Get the content to import
            fp = open(item["filepath"], 'r')
            new_content = fp.read()
            fp.close()
            author = entry.get("author") or self.default_author
            if item["object"] is None and self.smart:
                item["object"] = Page(title=item["slug"], slug=item["slug"], author=author, content=new_content)
                if self.verbosity > 0:
                    print "Document '{0}' does not exist, create it".format(item["slug"])
            else:
                item["object"].author = author
                item["object"].content = new_content
                if self.verbosity > 0:
                    print "Document '{0}' updated".format(item["slug"])
                
            item["object"].comment = "Imported from file '{0}'".format(item["_filepath_filename"])
            
            if not self.dry_run:
                item["object"].save()

    def do_export(self, entry):
        """
        Export documents to their given file destination
        """
        if self.verbosity > 0:
            print "=== Export '{0}' ===".format(entry["name"])
            
        for i, item in enumerate(entry['items'], start=1):
            # Create the missing dir if smart mode is enabled
            if self.smart and not os.path.exists(item["_filepath_dir"]):
                if not self.dry_run:
                    os.makedirs(item["_filepath_dir"])
            # Create the file to its destination
            if not self.dry_run:
                fp = open(item["filepath"], 'w')
                fp.write(item['object'].content.encode('UTF8'))
                fp.close()
            if self.verbosity > 0:
                print "Document '{0}' writed to: {1}".format(item["slug"], item["filepath"])
